/*
 *	$Id: link.cpp 1.8 Sat, 06 Dec 2003 14:05:33 -0800 root $
 *	$ProjectHeader: pptpproxy 1.10 Sat, 06 Dec 2003 14:45:37 -0800 root $
 */

#include <proxy.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

Proxy::Link::Link(
Proxy	*_proxy,
int	pairIndex
)
{
    proxy= _proxy;

    callerSocket= -1;
    calleeSocket= -1;
    pair= proxy->pairs[pairIndex];

    callerIP= ~0;
    fakeCallerId= ~0;
    realCallerId= ~0;

    calleeIP= ~0;
    fakeCalleeId= ~0;
    realCalleeId= ~0;

    proxy->dbg(
	"%s: incoming connection on %s",
	__func__,
	pair->getListenName()
    );

    int s= pair->getSocket();
    struct sockaddr_in addr;
    socklen_t addrLen= sizeof(addr);
    s= accept(s,(struct sockaddr*)&addr,&addrLen);
    if(s<0)
    {
	proxy->fail(
	    false,
	    "accept",
	    "%s: accept failed on pair %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
	return;
    }

    if(proxy->isDebugOn())
    {
	proxy->dbg(
	    "%s: accepted connection from %s",
	    __func__,
	    inet_ntoa(addr.sin_addr)
	);
    }

    if(proxy->setNonBlocking(s,true,false)==false)
    {
	proxy->fail(
	    false,
	    0,
	    "%s: setNonBlocking failed on pair %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
	return;
    }

    IPAddr inIP= (IPAddr)addr.sin_addr.s_addr;
    if(proxy->checkACL(inIP)==false)
    {
	proxy->fail(
	    false,
	    0,
	    "%s: unauthorized connection from IP %s",
	    __func__,
	    inet_ntoa(addr.sin_addr)
	);
	close(s);
	return;
    }
    callerSocket= s;
    callerIP= inIP;

    if(proxy->isDebugOn())
    {
	proxy->dbg(
	    "%s: connecting to peer %s",
	    __func__,
	    pair->getPeerName()
	);
    }

    memset(&addr,0,sizeof(addr));
    addr.sin_family= AF_INET;
    addr.sin_port= (uint16_t)pair->getPeerPort();
    addr.sin_addr.s_addr= pair->getPeerAddr();
    s= proxy->makeSocket(SOCK_STREAM,0,false);
    if(connect(s,(struct sockaddr*)&addr,sizeof(addr))<0)
    {
	proxy->fail(
	    false,
	    "connect",
	    "%s: failed to connect to server %s",
	    __func__,
	    pair->getPeerName()
	);
	close(callerSocket);
	callerSocket= -1;
	close(s);
	return;
    }

    if(proxy->setNonBlocking(s,true,false)==false)
    {
	proxy->fail(
	    false,
	    "connect",
	    "%s: setNonBlocking failed on pair %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
	close(callerSocket);
	callerSocket= -1;
	close(s);
	return;
    }
    calleeSocket= s;
    calleeIP= pair->getPeerAddr();

    if(proxy->isDebugOn())
    {
	proxy->dbg(
	    "%s: tcp link established %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
    }
}

Proxy::Link::~Link()
{
    if(proxy->isDebugOn())
    {
	if(
	    callerSocket>=0 &&
	    calleeSocket>=0
	)
	{
	    char *srcName= strdup(proxy->makeString(callerIP));
	    char *dstName= strdup(proxy->makeString(calleeIP));
	    proxy->dbg(
		"%s: shutting down link %s -> %s",
		__func__,
		srcName,
		dstName
	    );
	    free(srcName);
	    free(dstName);
	}
	else
	{
	    proxy->dbg(
		"%s: destroying invalid link (unauthorized)",
		__func__
	    );
	}
    }
    close(callerSocket);
    close(calleeSocket);
}

bool Proxy::Link::tcpPacket(
bool callerPacket
)
{
    int src= getCalleeSocket();
    int dst= getCallerSocket();
    if(callerPacket)
    {
	int t= src;
	src= dst;
	dst= t;
    }

    uint8_t buf[16384];
    int n= read(src,buf,16384);
    if(n==0)
    {
	proxy->fail(	
	    false,
	    0,
	    "%s: EOF condition on control socket for pair %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
	return false;
    }
    else if(n<0)
    {
	if(errno==EINTR) return true;
	if(errno==EAGAIN) return true;
	proxy->fail(	
	    false,
	    "read",
	    "%s: tcp read failed on control socket for pair %s -> %s",
	    __func__,
	    pair->getListenName(),
	    pair->getPeerName()
	);
	return false;
    }

    if(proxy->isDebugOn())
    {
	char *calleeName= strdup(proxy->makeString(calleeIP));
	char *callerName= strdup(proxy->makeString(callerIP));
	proxy->dbg(
	    "%s: incoming tcp packet, type 0x%x on link %s %s %s",
	    __func__,
	    buf[9],
	    callerName,
	    callerPacket ? "->" : "<-",
	    calleeName
	);
	free(callerName);
	free(calleeName);
    }

    if(
	buf[2]==0x00	&&	// Control packet
	buf[3]==0x01	&&
	buf[4]==0x1A	&&	// Magic
	buf[5]==0x2B	&&
	buf[6]==0x3C	&&
	buf[7]==0x4D
    )
    {
	if(
	    buf[8]==0x00	&&	// Call-Request
	    (
		buf[9]==0x07	||
		buf[9]==0x08
	    )
	)
	{
	    uint32_t fakeId;
	    uint32_t id= buf[12] | (((uint16_t)buf[13])<<8);
	    uint32_t serial= buf[14] | (((uint16_t)buf[15])<<8);
	    if(callerPacket)
	    {
		realCallerId= id;
		fakeCallerId= fakeId= proxy->allocCallerId();
	    }
	    else
	    {
		realCalleeId= id;
		fakeCalleeId= fakeId= proxy->allocCalleeId();
	    }
	    buf[12]= (fakeId>>0)&0xFF;
	    buf[13]= (fakeId>>8)&0xFF;

	    proxy->dbg(
		"%s: tcp packet realCallId= 0x%x, fakeCallId= 0x%x %s= 0x%x",
		__func__,
		id,
		fakeId,
		buf[9]==0x08 ? "peerId" : "serial",
		serial
	    );

	    if(buf[9]==0x08)
	    {
		CallId peer= realCallerId;
		buf[14]= (peer>>0)&0xFF;
		buf[15]= (peer>>8)&0xFF;
	    }
	}
	else if(
	    buf[8]==0x00	&&
	    (
		buf[9]==0x0C	||	// Call-Clear-Request
		buf[9]==0x0D	||	// Call-Disconnect-Notify
		buf[9]==0x0E	||	// WAN-Error-Notify
		buf[9]==0x0F		// Set-Link_Info
	    )
	)
	{
	    uint32_t mappedId;
	    uint32_t id= buf[12] | (((uint16_t)buf[13])<<8);
	    if(proxy->remapId(&mappedId,id))
	    {
		buf[12]= (mappedId>>0)&0xFF;
		buf[13]= (mappedId>>8)&0xFF;
	    }
	}
    }

    proxy->dbg(
	"%s: forwarding tcp packet",
	__func__
    );

    uint8_t *p= buf;
    while(n>0)
    {
	int s= write(dst,p,n);
	if(s==0)
	{
	    proxy->fail(	
		false,
		0,
		"%s: wrote 0 bytes on control socket for pair %s -> %s",
		__func__,
		pair->getListenName(),
		pair->getPeerName()
	    );
	    return false;
	}
	else if(s>0)
	{
	    p+= s;
	    n-= s;
	}
	else
	{
	    if(errno==EINTR)		continue;
	    else if(errno==EAGAIN)	continue;
	    else
	    {
		if(proxy->isDebugOn())
		{
		    proxy->fail(	
			false,
			"write",
			"%s: tcp write failed on control socket for pair %s -> %s",
			__func__,
			pair->getListenName(),
			pair->getPeerName()
		    );
		}
		return false;
	    }
	}
    }

    return true;
}

